
#ifndef CALLBACK_PTR_HPP
#define CALLBACK_PTR_HPP

#include <functional>
#include <memory>
#include <mutex>

#include <QPointer>
#include <QThread>

// Internal struct with data shared between safe_callback and safe_callback_ctrl
struct safe_callback_data
{
    std::mutex mutex;
    bool callbackAlive;

    safe_callback_data(): mutex(), callbackAlive(true) {}
};

// safe_callback is generated by safe_callback_ctrl.
// it will be a valid functor as long as safe_callback_ctrl exists.
// When safe_callback_ctrl is destroyed, all its safe_callbacks will become
// empty. operator() will do nothing.
// Usefull when calling object's method from another thread
// as we are sure target object exists (or nothing happens)
template<typename... Args>
class safe_callback final
{
    public:
        safe_callback(const std::shared_ptr<safe_callback_data>& data, const std::function<void(Args...)>& callback): m_data(data), m_callback(callback) {}
        safe_callback(const safe_callback<Args...> &) = default;

        safe_callback& operator=(const safe_callback<Args...> &) = default;

        virtual ~safe_callback() = default;

        void operator()(Args... args) const
        {
            std::lock_guard<std::mutex> lock(m_data->mutex);

            if (is_valid())
                m_callback(std::forward<Args>(args)...);
        }

        bool is_valid() const
        {
            return m_data->callbackAlive == true && m_callback;
        }

    private:
        std::shared_ptr<safe_callback_data> m_data;

        std::function<void(Args...)> m_callback;
};


// safe_callback_ctrl controls and creates safe_callbacks
class safe_callback_ctrl final
{
    public:
        safe_callback_ctrl(): m_data()
        {
            setup();
        }

        safe_callback_ctrl(const safe_callback_ctrl &) = delete;

        ~safe_callback_ctrl()
        {
            // safe_callback_ctrl should be reset before its destruction.
            //
            // Object owning instance of this class should call invalidate()
            // as a first action in its destructor.
            // Otherwise owner class may be partialy destructed before
            // safe_callback_ctrl is reseted which will allow
            // callbacks to work on a broken object
            assert(m_data.get() == 0 || m_data.use_count() == 1);  // no data, or only one client - us

            reset();
        }

        template<typename... R, typename T>
        auto make_safe_callback(const T& callback) const
        {
            safe_callback<R...> callbackPtr(m_data, callback);

            return callbackPtr;
        }

        safe_callback_ctrl& operator=(const safe_callback_ctrl &) = delete;

        void invalidate()
        {
            reset();      // dissolve all connections
            setup();      // create new one
        }

    private:
        template<typename...>
        friend class safe_callback;

        std::shared_ptr<safe_callback_data> m_data;

        void setup()
        {
            m_data = std::make_shared<safe_callback_data>();
        }

        void reset()
        {
            // mark all safe callbacks as invalid
            {
                // lock resource
                std::lock_guard<std::mutex> lock(m_data->mutex);

                // mark resource as dead
                m_data->callbackAlive = false;
            }

            // detach from existing safe callbacks
            m_data.reset();
        }
};


// extends QMetaObject::invokeMethod by version with arguments
template<typename Obj, typename F, typename... Args>
void invokeMethod(Obj* object, const F& method, Args&&... args) requires std::is_base_of<QObject, Obj>::value
{
    QMetaObject::invokeMethod(object, [object, method, args...]()
    {
        (object->*method)(args...);
    });
}


// call_from_object_thread uses Qt mechanisms to invoke function in another thread
// (thread of 'object' object)
template<typename F, typename ObjT, typename... Args>
requires std::is_base_of_v<QObject, ObjT>
void call_from_object_thread(QPointer<ObjT> object, const F& function, Args&&... args)
{
    if (object.data() != nullptr)
    {
        if constexpr (std::is_member_function_pointer_v<F>)
            QMetaObject::invokeMethod(object.data(), [object, function, args...]()
            {
                ((*object).*function)(args...);
            });
        else
            QMetaObject::invokeMethod(object.data(), [function, args...]()
            {
                function(args...);
            });
    }
}


template<typename F, typename... Args>
void call_from_this_thread(QThread* thread, const F& function, Args&&... args)
{
    QMetaObject::invokeMethod(thread, [function, args...]()
    {
        function(args...);
    });
}


// construct a functor which invoked will invoke encapsulated
// functor in another thread
template<typename... Args, typename ObjT, typename F>
requires std::is_base_of_v<QObject, ObjT>
std::function<void(Args...)> make_cross_thread_function(ObjT* object, const F& function)
{
    std::function<void(Args...)> result = [=](Args&&... args)
    {
        call_from_object_thread(QPointer<ObjT>(object), function, std::forward<Args>(args)...);
    };

    return result;
}


template<typename... Args, typename F>
std::function<void(Args...)> make_cross_thread_function(QThread* thread, const F& function)
{
    std::function<void(Args...)> result = [=](Args&&... args)
    {
        call_from_this_thread(thread, function, std::forward<Args>(args)...);
    };

    return result;
}


// construct a functor which invoked will invoke a method
// (slot) of given object. Will do nothing when given object is destroyed.
// Similar to safe_callback_ctrl (but method will be invoked in target's thread)
template<typename ObjT, typename R, typename ...Args>
requires std::is_base_of_v<QObject, ObjT>
auto queued_slot(ObjT* obj, R(ObjT::*method)(Args...))
{
    QPointer<ObjT> objPtr(obj);

    return [objPtr, method](Args... args)
    {
        ObjT* object = objPtr.data();

        if (object)
            invokeMethod(object, method, std::forward<Args>(args)...);
    };
}

// construct a functor which invoked will invoke a method
// (slot) of given object. Will do nothing when given object is destroyed.
// Similar to safe_callback_ctrl (but uses Qt mechanism to guarantee (?) threadsafety)
// In contrast to queued_slot() method is invoked in caller's thread
template<typename ObjT, typename R, typename ...Args>
requires std::is_base_of_v<QObject, ObjT>
std::function<void(Args...)> direct_slot(ObjT* obj, R(ObjT::*method)(Args...))
{
    QPointer<ObjT> objPtr(obj);

    return [objPtr, method](Args... args)
    {
        ObjT* object = objPtr.data();

        if (object)
            (object->*method)(std::forward<Args>(args)...);
    };
}

#endif
